#   Compatibility with python 3.x
# from __future__ import print_function
# from __future__ import unicode_literals
# from __future__ import division

AUTHOR = 'Tomas Psorn'
CONTACT = 'tomaspsorn@isibrno.cz'

import math
import sys
import collections
import time

import matplotlib.pyplot as plt
import numpy as np

from isinmrtbx.tools.utils import combineChannels


def flashReco(dataObject):

    kspace = dataObject.fid

    dataOut = np.fft.fftshift(np.fft.ifft2(kspace, axes=(0,1)), axes=(0,1))

    return dataOut

"""
EPI reco
"""

def epiReco(dataObject):
    """
    EPI Reco pipeline

    Working with navigator data
    ~ Mirror odd lines
    ~ Do FFT in read direction (step into hybrid space)
    ~ Get phase of adjacent lines in each segments
    ~ Unwrap and fit the phase to a line and get slope and offset

    Working with data
    ~ Every other line is corrected by subtracting the phase slope
    ~ All measured k-space data distributed to a zero, reco-sized matrix
    ~ Grappa reco is used to calculate missing lines if any
    ~ Partial Fourier reco is used to correct for k-space truncation if any
    ~ FFT in phase direction is done
    ~ Multifrequency reco is done to correct for B0 inhomogenities

    Parameters
    ----------
    dataObject

    Returns
    -------

    """


    dimXr = dataObject.method.PVM_EncMatrix[0]  # number of pixels in read direction in the resulting image
    dimYr = dataObject.method.PVM_EncMatrix[1]  # number of phase encoding lines in the raw data.
    dimXi = dataObject.method.PVM_Matrix[0]     # number of samples in read direction in the raw data
    dimYi = dataObject.method.PVM_Matrix[1]     # number of pixels in the PE direction in the resulting image
    dimZ = dataObject.acqp.NI   # number of slices
    dimRep = dataObject.acqp.NR     #number of repetitions
    dimSeg = dataObject.method.NSegments    # number of segments
    dimCh = dataObject.getSelectedReceivers()       # number of receivers
    dimYrSeg = dimYr / dimSeg

    omNavigators_nr = None
    phaseCorr = np.zeros((dimCh, dimYrSeg , dimXr, dimZ * dimSeg), dtype='f4') # An array to store the phase correction information of every data point.

    # if there are navigators measured, use them for phase correction, if not, use Bruker stuff
    try:
        omNavigators_nr = dataObject.method.OMNavigators    # how many navigators were measured
        dimRep -= omNavigators_nr   # navigators are stored in a dimension together with repetitions
    except:
        pass

    if omNavigators_nr:
        navData = dataObject.fid[:, :, :, 0: dimZ * omNavigators_nr * dimSeg]  # split fid to navigators and data
        rawData = dataObject.fid[:, :, :, dimZ * omNavigators_nr * dimSeg:]

        demo_interSegmentPhaseCorr(navData, rawData, dataObject)
        sys.exit(1)

        # calculate phase correction for each channel
        for _dimCh in range(0, dimCh):
            phaseCorr[_dimCh,:,:,:] = calcPhaseCorEpi(dataObject, navigator=navData[_dimCh,:,:,:])  # it'll just read the bruker slopes and offsets
            #todo Intensity correction
    else:
        rawData = dataObject.fid  # there are no navigators i.e. all fid is data
        #todo: do this



    resultChan = np.zeros((dimCh, dimYi, dimXi, dimZ), dtype=rawData.dtype) # array to store reco results for each channel
    resultRep = np.zeros((dimYi, dimXi, dimZ, dimRep), dtype=rawData.dtype) # array to store reco results for each repetition

    # reco goes volume by volume, since some intra volume correction might be used eventually
    for _dimRep in range(0, dimRep):
        for _dimCh in range(0, dimCh):
            # get the particular volume and do the reco
            volume = rawData[_dimCh,:,:, _dimRep * dimZ * dimSeg : (_dimRep+1) * dimZ * dimSeg]
            temp = epiRecoRaw(np.squeeze(navData), dataObject, phaseCorr[_dimCh, :, :, :])
            resultChan[_dimCh, :, :, :] = epiRecoRaw (volume, dataObject, phaseCorr[_dimCh,:,:,:])

        resultRep[:,:,:, _dimRep] = combineChannels(resultChan)


    return resultRep



def epiRecoRaw (kspace, dataObject, phaseCorr):

    XdimI = dataObject.method.PVM_Matrix[0] # image size
    YdimI = dataObject.method.PVM_Matrix[1]
    segments_nr = dataObject.method.NSegments

    # kspace = intensityCorEpi(kspace, intensityCor)

    kspace = utils.mirrorOddLines(kspace, axis=1)

    # for i in range(0,20):
    #     plt.figure()
    #     plt.imshow(phaseCorr[:,:,i], interpolation='None')
    #     plt.show()

    plt.figure()
    plt.subplot(1,2,1)
    plt.imshow(np.abs(kspace[:,:,0]), interpolation='None')

    hybrid = np.fft.ifft2(kspace, axes=(1,))

    hybrid = np.fft.fftshift(hybrid, axes=1)

    hybrid =  phaseCorEpi(hybrid, phaseCorr)

    hybrid = sortEpi(hybrid, dataObject)

    kspace = np.fft.fft2(hybrid, axes=(1,))

    plt.subplot(1, 2, 2)
    plt.imshow(np.log(np.abs(kspace[:, :, 0])), interpolation='None')
    plt.show()

    image = (np.fft.ifft2(hybrid, s=(YdimI,), axes=(0,)))

    image = np.fft.fftshift(image, axes=0)

    for i in range(0,2):
        plt.figure()
        plt.imshow(np.log(np.abs(kspace[:, :, i])), interpolation='None')
        plt.colorbar()
        plt.show(    )

    return image



def calcPhaseCorEpi(dataObject, **kwargs):
    """
    Function to get phase correction coefficients.
    This can be done in two ways.

    1/  Navigator scans are measured
    2/  If navigator scans are not a part of the dataset,
        then the Bruker calculated coefficients are used

    Parameters
    ----------
    dataObject
    kwargs

    Returns
    -------
    slopes
     - dim(lines per segment, )

    """
    dimXr = dataObject.method.PVM_EncMatrix[0]  # number of pixels in read direction in the resulting image
    dimYr = dataObject.method.PVM_EncMatrix[1]  # number of phase encoding lines in the raw data.
    dimZ = dataObject.acqp.NI  # number of slices
    dimSeg = dataObject.method.NSegments  # number of segments
    dimHigh = dimZ * dimSeg
    dimYrSeg = int (dimYr / dimSeg)

    navData = None

    phaseCorr = np.zeros((dimYrSeg, dimXr, dimHigh), dtype='f4')  # An array to store the phase correction information of every data point.


    # test if navigator scan is present
    try:
        navData = kwargs['navigator']
        mode = 'nav'
    except:
        mode = 'obj'



    # calculate phase correction coefficients
    if mode == 'nav':

        phaseFirstSegEven = np.zeros ((dimYrSeg / 2, dimXr), dtype='f4')     # stores the phase of every even line of the first segment
        phaseFirstSegOdd = np.zeros ((dimYrSeg/ 2, dimXr), dtype='f4')       # stores the phase of every odd line of the first segment

        slopesOddEven = np.zeros ((dimYrSeg / 2, dimHigh), dtype='f4')
        offsetsOddEven = np.zeros ((dimYrSeg / 2, dimHigh), dtype='f4')

        slopesInterSegEven = np.zeros ((dimYrSeg / 2, dimHigh), dtype='f4')
        offsetsInterSegEven = np.zeros ((dimYrSeg / 2, dimHigh), dtype='f4')

        slopesInterSegOdd = np.zeros ((dimYrSeg / 2, dimHigh), dtype='f4')
        offsetsInterSegOdd = np.zeros ((dimYrSeg / 2, dimHigh), dtype='f4')


        # mirror odd lines
        navData = utils.mirrorOddLines(navData,1)
        # FFT in read dirrection

        navData = np.fft.fftshift(np.fft.ifft2(navData, axes=(1,)), axes=1)

        temp = np.unwrap(np.angle(navData[0,:,0:10]), axis=0)
        temp2 = np.unwrap(np.angle(navData[0,:,10:20]), axis=0)

        plt.figure()
        plt.suptitle('Segment to segment phase change as it goes trough the readout', fontsize=18)
        for i in range(28,68,5):
            plt.subplot(1,2,1)
            plt.plot(temp[i,:], label='sample = %s' % i)
            plt.subplot(1, 2, 2)
            plt.plot(temp2[i, :], label='sample = %s' % i)

        plt.subplot(1, 2, 1)
        plt.title("Slice nr. 1")
        plt.xlabel('Segment nr.')
        plt.ylabel('Phase [rad]')

        plt.subplot(1, 2, 2)
        plt.title("Slice nr. 2")
        plt.xlabel('Segment nr.')
        plt.ylabel('Phase [rad]')

        plt.legend()

        plt.figure()
        for i in range(0, 10):
            plt.plot(temp[:, i], label='segment = %s' % i)

        plt.xlabel('x')
        plt.ylabel('Phase [rad]')
        plt.title('Segment to segment phase change in different stages of readout')
        plt.legend()


        plt.show()

        sys.exit(1)


        for _dimHigh in range(0, dimHigh):

            valid_points = utils.markValidPoints1(navData[dimYrSeg/2, :, _dimHigh]) # the middle line

            for _dimYrSeg in range(0, dimYrSeg, 2):

                phaseEven = np.angle(navData[_dimYrSeg, :, _dimHigh])
                phaseOdd = np.angle(navData[_dimYrSeg + 1, :, _dimHigh])

                print(dimYrSeg)

                # get a reference for a intersegment correctu
                if (_dimHigh == 0):    # for every line of first segment
                    phaseFirstSegEven[_dimYrSeg / 2, :] = phaseEven
                    phaseFirstSegOdd[_dimYrSeg / 2, :] = phaseOdd
                    print("haha")

                # get phase difference between adjacent lines
                phaseOddEven = phaseOdd - phaseEven
                phaseIntersegEven = phaseEven - phaseFirstSegEven[_dimYrSeg / 2,:]
                phaseIntersegOdd = phaseOdd - phaseFirstSegOdd[_dimYrSeg / 2,:]

                # plt.figure()
                # plt.subplot(3,3,1)
                # plt.plot(phaseOddEven)
                # plt.subplot(3,3,2)
                # plt.plot(phaseIntersegEven)
                # plt.subplot(3,3,3)
                # plt.plot(phaseIntersegOdd)

                phaseOddEven *= valid_points
                phaseIntersegEven *= valid_points
                phaseIntersegOdd *= valid_points

                # plt.subplot(3,3,4)
                # plt.plot(phaseOddEven)
                # plt.subplot(3,3,5)
                # plt.plot(phaseIntersegEven)
                # plt.subplot(3,3,6)
                # plt.plot(phaseIntersegOdd)


                # unwrap phase
                phaseOddEven = np.unwrap(phaseOddEven)
                phaseIntersegEven = np.unwrap(phaseIntersegEven)
                phaseIntersegOdd = np.unwrap(phaseIntersegOdd)

                phaseOddEven *= valid_points
                phaseIntersegEven *= valid_points
                phaseIntersegOdd *= valid_points

                # plt.subplot(3,3,7)
                # plt.plot(phaseOddEven)
                # plt.subplot(3,3,8)
                # plt.plot(phaseIntersegEven)
                # plt.subplot(3,3,9)
                # plt.plot(phaseIntersegOdd)
                # plt.show()

                # non-valid points are set to zero

                offset, slope, offset_sd, slope_sd = utils.linearFitSafe(phaseOddEven)
                offsetsOddEven[_dimYrSeg / 2, _dimHigh] = offset
                slopesOddEven [_dimYrSeg / 2, _dimHigh] = slope

                offset, slope, offset_sd, slope_sd = utils.linearFitSafe(phaseIntersegEven)
                offsetsInterSegEven [_dimYrSeg / 2, _dimHigh] = offset
                slopesInterSegEven [_dimYrSeg / 2, _dimHigh] = slope

                offset, slope, offset_sd, slope_sd = utils.linearFitSafe(phaseIntersegOdd)
                offsetsInterSegOdd [_dimYrSeg / 2, _dimHigh] = offset
                slopesInterSegOdd [_dimYrSeg / 2, _dimHigh] = slope

        # correct by fit over slices

        for _dimHigh in range(0, dimHigh):

            for _dimYrSeg in range(0, dimYrSeg):

                if int(_dimYrSeg % 2) == 0:
                    line = utils.calcLine(dimXr, slopesInterSegEven[_dimYrSeg / 2, _dimHigh], offsetsInterSegEven[_dimYrSeg / 2, _dimHigh])
                    phaseCorr[_dimYrSeg, :, _dimHigh] += line

                else:
                    line = utils.calcLine (dimXr, slopesOddEven[_dimYrSeg / 2, _dimHigh], offsetsOddEven[_dimYrSeg / 2, _dimHigh])
                    line += utils.calcLine(dimXr, slopesInterSegOdd[_dimYrSeg / 2, _dimHigh], offsetsInterSegOdd[_dimYrSeg / 2, _dimHigh])
                    phaseCorr[_dimYrSeg, :, _dimHigh] += line



    #get bruker correction coefficients
    if mode == 'obj':

        # get an idea about the dimensionality of the dataset
        NSegments = dataObject.method.NSegments
        NI = dataObject.acqp.NI
        dim0 = int(dataObject.method.PVM_EncMatrix[1] / NSegments)
        dim1 = dataObject.method.PVM_EncMatrix[0]
        dim2 = int(NI * NSegments)

        slopes = np.zeros((dim0 / 2, dim2), dtype='f4', order='C')
        offsets = np.zeros((dim0 / 2, dim2), dtype='f4', order='C')

        phaseCor = dataObject.method.PVM_EpiPhaseCorrection

        for z in range(0, dim2):
            for line in range(0, dim0, 2):
                offsets[line / 2, z] = phaseCor [0, int(z/NSegments)+2] #todo finish
                slopes[line / 2, z] = phaseCor [0, int(z/NSegments)]

    return phaseCorr


def calcIntensityCorEpi(dataObject, **kwargs):

    NI = dataObject.acqp.NI
    NSegments = dataObject.method.NSegments

    intensityCor = np.zeros((NSegments, NI), dtype='f4')

    dim2 = NI * NSegments

    try:
        navigator = kwargs['navigator']
        mode = 'nav'
    except:
        mode = 'obj'
        pass



    if mode == 'nav':

        for _dim2 in range(0,dim2):
            _frame, _segment = divmod(_dim2, NSegments)

            segment = navigator[:,:, (_frame * NSegments) + _segment]

            intensityCor[_segment, _frame] = np.sum(np.abs(segment))

        for _frame in range(0, NI):
            intensityCor[:,_frame] /= np.amax(intensityCor[:,_frame])
            pass

    else:
        pass

    return intensityCor


def phaseCorEpi(dataIn, phaseCorr):

    linesPerSeg = dataIn.shape[0]
    dimXr = dataIn.shape[1]
    dimHigh = dataIn.shape[2]

    dataOut = np.zeros_like(dataIn)


    for _dimHigh in range(0, dimHigh):
        for _linesPerSeg in range(0, linesPerSeg):
            for _dimXr in range(0, dimXr):
                phase = np.complex(0.0, phaseCorr[_linesPerSeg , _dimXr , _dimHigh])

                dataOut[_linesPerSeg , _dimXr , _dimHigh] = dataIn[_linesPerSeg , _dimXr , _dimHigh] * np.exp(-1.0 * phase)

    return dataOut

def intensityCorEpi(dataIn, intensityCor):

    NSegments = intensityCor.shape[0]
    NI = intensityCor.shape[1]

    dim2 = dataIn.shape[2]

    for _dim2 in range(0, dim2):
        _frame, _segment = divmod(_dim2, NSegments)
        dataIn[:, :, (_frame * NSegments) + _segment] /= intensityCor[_segment,_frame]

    return dataIn


def prepareMultifrequencyReco(brukerObj, numberOfColumns, numberOfPElines, swh):

    # Get esential parameters from bruker data object
    fieldMap = brukerObj.data2dseq.data


    # Derive rest of parameters
    dwellTime_ms = 1.0 / float(swh)
    print (dwellTime_ms)
    acqTimeLine = dwellTime_ms * numberOfColumns
    minFieldMap = np.amin(fieldMap)
    maxFieldMap = np.amax(fieldMap)
    bandwidthPerVoxelPE = 1.0/acqTimeLine/numberOfPElines
    numberOfFrequencies = int(2.0 * (maxFieldMap - minFieldMap)/bandwidthPerVoxelPE)
    frequencyStep = (maxFieldMap - minFieldMap)/numberOfFrequencies

    # Alloc array to store all the offset frequencies
    frequenciesArray = np.zeros(frequencyStep, dtype=minFieldMap.dtype, order='F')

    # Fill frequenciesArray
    for i in range(0,numberOfFrequencies):
        frequenciesArray[i] = minFieldMap + i * frequencyStep + .5 * frequencyStep

    return frequenciesArray

def sortEpi(dataIn, dataObject):

    dimSeg = dataObject.method.NSegments
    dimZ = dataObject.acqp.NI

    PVM_EncSteps1 = np.asarray(dataObject.method.PVM_EncSteps1)
    PVM_EncSteps1 += np.abs(np.amin(PVM_EncSteps1))

    dimYrSeg = dataIn.shape[0]
    dimXr = dataIn.shape[1]
    dimHigh = dataIn.shape[2]

    dimYr = dimYrSeg * dimSeg

    dataOut = np.zeros((dimYr,dimXr, dimZ), dtype=dataIn.dtype)

    for _dimHigh in range(0, dimHigh):

        _dimZ = int(_dimHigh / dimSeg)
        _dimSeg = _dimHigh % dimSeg

        for _dimYrSeg in range(0, dimYrSeg):

            _PVM_EncSteps1 = _dimSeg *  dimYrSeg + _dimYrSeg
            _dataOut = PVM_EncSteps1[_PVM_EncSteps1]
            dataOut[_dataOut, :, _dimZ] = dataIn[_dimYrSeg, :, _dimHigh]

    return dataOut

"""
Gridding stuff
"""

def gridding_dcf1(traj, fog, kerWidth, iter_nr=5):

    out = np.ones_like(traj,order='F')

    for _iter in range(0,iter_nr):

        grid_tmp = gridding1_1(out, traj, fog, kerWidth, trim=True, doDcf=False)

        weights_tmp = degridding1_1(grid_tmp, traj, fog, kerWidth)

        for i in range(0,len(weights_tmp)):
            if weights_tmp[i] == 0.: out[i] = 0.
            else: out[i] /= weights_tmp[i]


    return out


def gridding1_1 (dataIn, trajIn, fog, kernelWid_n, trim=True, doDcf = True):
    """
    _n - points
    _k - kspace
    _g - gridspace
    """

    if doDcf:
        iterations = 4
        densityCompensationFactors = gridding_dcf1(trajIn, fog, kernelWid_n, iterations)
        dataIn *= densityCompensationFactors

    kernelSamples_nr = 1000
    beta = np.pi*np.sqrt( math.pow(kernelWid_n,2)/math.pow(fog,2)*math.pow((fog-0.5),2)-0.8 )
    kernel = np.kaiser(2 * kernelSamples_nr, beta)[-kernelSamples_nr:]

    dim0_n = len(dataIn)
    dim0_g = dim0_n * fog

    kernelWid_k = kernelWid_n / float(dim0_n)
    kernelWid_k_div2 = kernelWid_k / 2.

    ddim0_k = 1.0 / float(dim0_n)
    ddim0_g = 1.0/ float(dim0_g)

    kmax = 0.5 + kernelWid_k_div2
    overlap = np.ceil(kernelWid_k_div2 / ddim0_g)
    dataOut = np.zeros(dim0_g + 2 * overlap, dtype=dataIn.dtype, order='F')

    for data_x in range(0, dim0_n):
        data = dataIn[data_x]
        kx = trajIn[data_x]

        for grid_x in range(0, len(dataOut)):

            gx = grid_x * ddim0_g - kmax
            dx = abs(kx - gx)

            if dx < kernelWid_k_div2:
                ind = np.floor(kernelSamples_nr* dx / kernelWid_k_div2)
                ker = kernel[ind]
                dataOut[grid_x] += data * ker

    if trim:
        # Add overlaping parts
        dataOut[-2*overlap:-overlap] += dataOut[0:overlap]
        dataOut[overlap: 2*overlap] += dataOut[-overlap:]
        # Trim
        dataOut = dataOut[overlap:-overlap]

    return dataOut

def degridding1_1 (dataIn, trajIn, fog, kernelWid_n):
    """
    From grid coordinates to traj coordinates
    """

    kernelSamples_nr = 1000

    beta = np.pi * np.sqrt(math.pow(kernelWid_n, 2) / math.pow(fog, 2) * math.pow((fog - 0.5), 2) - 0.8)
    kernel = np.kaiser(2 * kernelSamples_nr, beta)[-kernelSamples_nr:]

    dim0_n = len(trajIn)
    dim0_g = len(dataIn)

    kernelWid_k_n = kernelWid_n / float(dim0_n)
    kernelWid_k_div2_n = kernelWid_k_n / 2.

    ddim0_g = (1 + kernelWid_k_n) / float(dim0_g)
    ddim0_k = 1 / float(dim0_n)

    kmax = 0.5 + kernelWid_k_div2_n

    dataOut = np.zeros(dim0_n, dtype=dataIn.dtype, order='F')



    for grid_x in range(0, dim0_g):

        gx = grid_x * ddim0_g - kmax
        data_g = dataIn[grid_x]

        for data_x in range(0, len(trajIn)):

            kx = trajIn[data_x]
            dx = abs(kx - gx)

            if dx < kernelWid_k_div2_n:
                ind = np.floor(kernelSamples_nr * dx / kernelWid_k_div2_n)
                ker = kernel[ind]
                dataOut[data_x] += data_g * ker

    return dataOut


def gridMinMax(x, maximum, radius):

    out = collections.namedtuple('MinMax',['min','max'])
    out.min = int(np.ceil(x-radius))
    out.max = int(np.floor(x+radius))
    if out.min < 0: out.min  = 0
    if out.max > maximum: out.max  = maximum

    return out

def gridMinMax1_1(x, maximum, radius):

    out = collections.namedtuple('MinMax',['min','max'])
    out.min = int(np.ceil(x-radius))
    out.max = int(np.floor(x+radius))
    if out.min < 0: out.min  = 0
    if out.max > maximum: out.max  = maximum

    return out

def _poly_sdc_kern_0lobes (r):

    POLY_ORDER = 5
    FIT_LEN = 394
    SPECTRAL_LEN = 25600
    FOV = 63

    x = SPECTRAL_LEN * r / float(FOV)

    poly = np.array ((-1.1469041640943728E-13,
                      8.5313956268885989E-11,
                      1.3282009203652969E-08,
                      -1.7986635886194154E-05,
                      3.4511129626832091E-05,
                      0.99992359966186584))

    out = poly[5]

    for i in range(1,POLY_ORDER+1):
        out += pow(x,i) * poly[POLY_ORDER-i]

    if out < 0:
        out = 0

    return out

def loadKernelTable(length):

    rfp = 0.96960938
    out = np.zeros(length,order='F')

    for i in range(0,length):
        out[i] = _poly_sdc_kern_0lobes(math.sqrt(rfp**2 * float(i) / float(length-1)))

    return out


def demo_interSegmentPhaseCorr (navData, rawData, dataObject):

    navData = np.squeeze(navData)

    navData = utils.mirrorOddLines(navData,1)

    navData = np.fft.fftshift(np.fft.ifft2(navData, axes=(1,)), axes=1)

    navData = np.unwrap(np.angle(navData), axis=1)

    print(navData.shape)

    plt.figure()
    plt.suptitle('Segment to segment phase change as it goes trough the readout', fontsize=18)
    for i in range(28, 68, 5):
        plt.subplot(1, 2, 1)
        plt.plot(navData[0, i, 0:10], label='sample = %s' % i)
        plt.subplot(1, 2, 2)
        plt.plot(navData[0, i, 10:20], label='sample = %s' % i)

    plt.subplot(1, 2, 1)
    plt.title("Slice nr. 1")
    plt.xlabel('Segment nr.')
    plt.ylabel('Phase [rad]')

    plt.subplot(1, 2, 2)
    plt.title("Slice nr. 2")
    plt.xlabel('Segment nr.')
    plt.ylabel('Phase [rad]')

    plt.legend()

    plt.figure()
    for i in range(0, 10):
        plt.plot(navData[0, :, 0:10], label='segment = %s' % i)

    plt.xlabel('x')
    plt.ylabel('Phase [rad]')
    plt.title('Segment to segment phase change in different stages of readout')
    plt.legend()

    plt.show()


    plt.figure()

    plt.show()

    return